local Dialog = LibStub("LibDialog-1.0")

local desireOptions = {"BiS", "Upgrade", "Off-spec", "Not Needed"}
local desireColors = {{r=0, g=1, b=0}, {r=1, g=1, b=0}, {r=1, g=0.5, b=0}, {r=1, g=0, b=0}}

local raidMembers = {}


Dialog:Register("BossListBlacklistConfirmationDialog", {
  width = 400,
  buttons = {
    {
      text = "Yes",
      on_click = function(self)
                    BossList.AddToBlackList(self.data.name)
                 end
    },
    {
      text = "No"
    }
  },
  on_show = function(self)
              self.text:SetText("You are about to add player |cFF3BFFF2" .. self.data.name .. "|r to your blacklist. By doing so, all entries of this player will be |cFFFF3B3Eremoved|r from BossList, and no entries from this player will be added again until this player has been removed from the blacklist through the interface options. Are you sure this is what you want?")
            end,
  hide_on_escape = true,
  
})

local function RaidChanged()
  wipe(raidMembers)
  if IsInRaid() then
    -- Lookup the raid members in the groups selected to be highlighted
    local highlightGroups = BossList.DB.global.highlightGroups
    for i = 1, GetNumGroupMembers() do
      local name, _, subgroup = GetRaidRosterInfo(i)
      
      if highlightGroups[subgroup] then --and name 
        raidMembers[name] = true
      end
    end
  end
  
  -- If the overview is currently being shown, update the inRaidHighlights
  if BossListScrollFrame:IsVisible() then
    -- Check if any of these members need anything
    for idx = 1, 4 do
      local desireList = _G["BossListDesire" .. idx]
      local entries = desireList.entries
      
      for i, entry in pairs(entries) do
        if entry:IsVisible() then
          if raidMembers[entry.name:GetText()] then
            entry.inRaidHighlight:Show()
          else
            entry.inRaidHighlight:Hide()
          end
        end
      end
    end
  end
end

BossListScrollFrame:SetScript("OnEvent", RaidChanged)
BossListScrollFrame:RegisterEvent("GROUP_ROSTER_UPDATE")

local function pairsByKeys (t, f)
	local a = {}
		for n in pairs(t) do table.insert(a, n) end
		table.sort(a, f)
		local i = 0      -- iterator variable
		local iter = function ()   -- iterator function
			i = i + 1
			if a[i] == nil then return nil
			else return a[i], t[a[i]]
			end
		end
	return iter
end

function BossList.DifficultyChanged(newDifficulty)
  -- If the new difficulty is not supported by BossList (== 0), disable the tab
  if newDifficulty == 0 then
    BossListJournalTab:Disable()
    
    -- Click the first tab if the BossList tab was selected
    if EncounterJournal.encounter.info.tab == 4 then
      EncounterJournalEncounterFrameInfoBossTab:Click()
    end
  else
    BossListJournalTab:Enable()
    -- If the BossList tab was being shown, make sure it is updated with the new difficulty
    if EncounterJournal.encounter.info.tab == 4 then
      BossListJournalTab:Click()
    end
  end
  
  -- Also update the list of raid members
  RaidChanged()
end

function BossList.BlacklistPlayer(name)
  Dialog:Spawn("BossListBlacklistConfirmationDialog", {name = name})
end

local function CreateDesireLists(frame)
  for idx, desire in pairs(desireOptions) do
    local desireList = CreateFrame("Frame", "BossListDesire" .. tostring(idx), frame, "BossListBossDesire")
    desireList.title:SetText(desire)
    desireList.title:SetTextColor(desireColors[idx].r, desireColors[idx].g, desireColors[idx].b)
    desireList.entries = {}
    
    if idx > 1 then
      desireList:SetPoint("TOP", "BossListDesire" .. tostring(idx - 1) .. "ListBG", "BOTTOM", 0, -5)
    end
  end
end

local createdDesireLists = false
function BossList.DisplayEntries()
  -- If it is the first time, setup the desire lists
  if not(createdDesireLists) then
    CreateDesireLists(BossListScrollFrameScrollChild)
    createdDesireLists = true
  end
  
  local instanceID = EncounterJournal.instanceID
  local encounterID = EncounterJournal.encounterID
  local difficultyID = BossList.difficulty
  
  -- Check if there are any entries for this instance-encounter-difficulty combination to show
  if BossList_entries[instanceID] == nil or 
     BossList_entries[instanceID][difficultyID] == nil or
     BossList_entries[instanceID][difficultyID][encounterID] == nil then
    
    -- Hide all entries
    for desireIdx, desire in pairs(desireOptions) do
      local desireList = _G["BossListDesire" .. desireIdx]
      local entries = desireList.entries
      for i = 1, #entries do
        entries[i]:Hide()
      end
      
      -- Retract the background
      desireList.listBG:SetPoint("BOTTOM", desireList, "BOTTOM")
    end
    
    return
  end
    
  local data = BossList_entries[instanceID][difficultyID][encounterID]
  
  -- Keep track of players that are already being shown, so they do not appear under multiple desires
  local printedPlayers = {}
  -- For each player, keep track of which items they need for the tooltip
  local playerItems = {}
  for desireIdx, desire in pairs(desireOptions) do
    local desireList = _G["BossListDesire" .. desireIdx]
    local entries = desireList.entries
    
    -- Find all players that have the current desire for this encounter, by checking each item
    local presentPlayers = {}
    for item, desires in pairs(data) do
      local players = desires[desireIdx] or {}
      
      for player, _ in pairs(players) do
        -- Check if the player should be printed in this desire category
        if not(printedPlayers[player]) then
          presentPlayers[player] = true
        end
        
        -- Record this player needs this item with this desire. Create tables if needed (Should change the datastructure at some point to make this duplication unneeded)
        if desireIdx < 4 then
          if not(playerItems[player]) then
            playerItems[player] = {}
          end
          if not(playerItems[player][desireIdx]) then
            playerItems[player][desireIdx] = {}
          end
          tinsert(playerItems[player][desireIdx], item)
        end
      end
    end
    
    -- Create entries for each of the players found
    local counter = 1
    for player, _ in pairsByKeys(presentPlayers) do
      local entry;
      if entries[counter] == nil then -- Create a new entry if needed
        entry = CreateFrame("Frame", "BossListEntryFrame", desireList, "BossListEntry")
        entry.name:SetTextColor(desireColors[desireIdx].r,desireColors[desireIdx].g, desireColors[desireIdx].b)
        
        -- Put it below the previous entry
        if counter > 1 then
          entry:SetPoint("TOP", entries[counter - 1], "BOTTOM")
        else
          entry:SetPoint("TOP", desireList, "BOTTOM", 0, -2)
        end
        
        entries[counter] = entry
      else
        entry = entries[counter]
        entry:Show()
      end
      
      entry.name:SetText(player)
      
      -- Show the inRaidHighlight if the player is currently in a group of the raid that fits in the current difficulty size
      if raidMembers[player] then
        entry.inRaidHighlight:Show()
      else
        entry.inRaidHighlight:Hide()
      end
      
      counter = counter + 1
      printedPlayers[player] = entry
    end
    
    -- Hide unused entries
    for i = counter, #entries do
      entries[i]:Hide()
    end
    
    -- Extend the background to the last entry, if any were added
    if counter > 1 then
      desireList.listBG:SetPoint("BOTTOM", entries[counter-1], "BOTTOM")
    else
      desireList.listBG:SetPoint("BOTTOM", desireList, "BOTTOM")
    end
  end
  
  -- Attach the item tables to the player entries
  for player, entry in pairs(printedPlayers) do 
    entry.items = playerItems[player] or {}
    
    -- Update the number of desired items indicators
    for idx = 1, 3 do
      local items = entry.items[idx]
      if items ~= nil then
        entry["desire" .. idx]:SetText(#items)
        entry["desire" .. idx]:Show()
      else
        entry["desire" .. idx]:Hide()
        --entry["desire" .. idx]:SetText(0)
      end
    end
  end
end

-- Enable/Disable the BossList tab depending on whether an instance is selected or not
hooksecurefunc("EncounterJournal_DisplayEncounter", function()
  if BossList.difficulty ~= 0 then
    BossListJournalTab:Enable()
  end
end)

hooksecurefunc("EncounterJournal_DisplayInstance", function()
  BossListJournalTab:Disable()
end)

-- Hook the hiding of the tab into the other tabs of the encounter journal
hooksecurefunc("EncounterJournal_TabClicked", function()
  BossListScrollFrame:Hide()
end)

-- The loottab button strangely enough doesn't call the global TabClicked function, so hook it directly
EncounterJournalEncounterFrameInfoLootTab:HookScript("OnClick", function()
  BossListScrollFrame:Hide()
end)

-- When DisplayEncounter is called, having the BossListTab selected causes an out of bounds error. Temporarily change the index when it is called.
local Old_EncounterJournal_DisplayEncounter = EncounterJournal_DisplayEncounter

EncounterJournal_DisplayEncounter = function(self, encounterID, noButton)
  local changedTab = false
  if EncounterJournal.encounter.info.tab == 5 then
    EncounterJournal.encounter.info.tab = 1
    changedTab = true
  end  
  
  Old_EncounterJournal_DisplayEncounter(self, encounterID, noButton)
  
  if changedTab then
    BossListJournalTab:Click()
  end
end

-- BossListItemsTooltip functions
function BossListItemsTooltip.SetOwner(self, owner, anchor, xOfs, yOfs)
  --self:SetParent(owner)
  self:ClearItems()
  self:ClearAllPoints()
  
  if anchor == "ANCHOR_TOP" then
    self:SetPoint("BOTTOM", owner, "TOP", xOfs, yOfs)
  elseif anchor == "ANCHOR_RIGHT" then
    self:SetPoint("BOTTOMLEFT", owner, "TOPRIGHT", xOfs, yOfs)
  elseif anchor == "ANCHOR_BOTTOM" then
    self:SetPoint("TOP", owner, "BOTTOM", xOfs, yOfs)
  elseif anchor == "ANCHOR_LEFT" then
    self:SetPoint("BOTTOMRIGHT", owner, "TOPLEFT", xOfs, yOfs)
  elseif anchor == "ANCHOR_TOPRIGHT" then
    self:SetPoint("BOTTOMRIGHT", owner, "TOPRIGHT", xOfs, yOfs)
  elseif anchor == "ANCHOR_BOTTOMRIGHT" then
    self:SetPoint("TOPLEFT", owner, "BOTTOMRIGHT", xOfs, yOfs)
  elseif anchor == "ANCHOR_TOPLEFT" then
    self:SetPoint("BOTTOMLEFT", owner, "TOPLEFT", xOfs, yOfs)
  elseif anchor == "ANCHOR_BOTTOMLEFT" then
    self:SetPoint("TOPRIGHT", owner, "BOTTOMLEFT", xOfs, yOfs)
  elseif anchor == "ANCHOR_CURSOR" then
    self:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", mX - (tooltipWidth / 2), mY)
  end
  
  for desire, items in pairs(owner.items) do
    if desire < 4 then
      for _, item in pairs(items) do
        self:AddItem(item, desire)
        
      end
    end
  end
end

function BossListItemsTooltip.ClearItems(self)
  for _, item in pairs(self.items) do
    item:Hide()
  end
  self.numItems = 0
end

function BossListItemsTooltip.AddItem(self, itemID, desire)
  local name, link, rarity, _, _, _, armorType, _, slot, icon, _ = GetItemInfo(itemID)
  local _, _, _, color = GetItemQualityColor(rarity)
  local numItems = self.numItems
  
  numItems = numItems + 1
  local itemEntry = self.items[numItems]
  if itemEntry == nil then
    
    itemEntry = CreateFrame("Button", "BossListItemsTooltipItem"  .. numItems, self, "BossListItemEntry")
    itemEntry.boss:Hide();
    itemEntry.bossTexture:Hide();
    itemEntry.bosslessTexture:Show();
    
    if numItems == 1 then
      itemEntry:SetPoint("TOPLEFT", self, 5, -5)
    else
      itemEntry:SetPoint("TOPLEFT", self.items[numItems - 1], "BOTTOMLEFT")
    end
    
    self.items[numItems] = itemEntry
  end
  
  itemEntry.name:SetText("|c" .. color .. name .. "|r");
  itemEntry.icon:SetTexture(icon);
  itemEntry.iconBorder:SetVertexColor(desireColors[desire].r, desireColors[desire].g, desireColors[desire].b)
  itemEntry.slot:SetText(_G[slot]);
  itemEntry.armorType:SetText(armorType);
  itemEntry.encounterID = EncounterJournal.encounterID;
  itemEntry.itemID = itemID;
  itemEntry.index = numItems;
  itemEntry.link = link;
  
  itemEntry:Show();
  
  self:SetHeight(45 * numItems + 10)
  
  self.numItems = numItems
end

function BossListItemsTooltip.Lock(self, entry)
  if self.lockedEntry then
    self.lockedEntry.highlight:SetAlpha(0)
  end
  
  self.locked = true  
  self.lockedEntry = entry
  self.lockedEntry.highlight:SetAlpha(1)
end

function BossListItemsTooltip.IsLocked(self)
  return self.locked
end

function BossListItemsTooltip.Unlock(self)
  if self.locked then
    if GetMouseFocus()  == self.lockedEntry then
      self.lockedEntry.highlight:SetAlpha(0.3)
    else
      self.lockedEntry.highlight:SetAlpha(0)
    end
    
    self.locked = false
    self.lockedEntry = nil
  end
end

local function AddPlayersGameTooltip(self)
  -- Check if data is present for the instance, difficulty, encounter, item combination
  local data = BossList_entries[EncounterJournal.instanceID]
  if data == nil then return; end
  
  local data = data[BossList.difficulty]
  if data == nil then return; end
  
  data = data[self.encounterID]
  if data == nil then return; end
  
  data = data[self.itemID]
  if data == nil then return; end
  
  local printedHeader = false
  for desire, players in pairs(data) do
    if desire < 4 and next(players) ~= nil then
      -- Print the header only if there is actually someone who wants this item
      if not(printedHeader) then
        GameTooltip:AddLine(" ")
        GameTooltip:AddLine("Desired by:")
        GameTooltip:AddTexture("Interface\\Buttons\\UI-GroupLoot-Dice-Up")
        printedHeader = true
      end 
      
      for player, _ in pairsByKeys(players) do
        GameTooltip:AddLine(player, desireColors[desire].r, desireColors[desire].g, desireColors[desire].b)
      end
    end
  end
  GameTooltip:Show()
end

function BossList.GroupSelector_OnClick(self, group)
  BossList.DB.global.highlightGroups[group] = not(BossList.DB.global.highlightGroups[group])
  RaidChanged()
end

-- Hook the GameTooltip modifying function into the EncounterJournal
for idx, item in pairs(EncounterJournal.encounter.info.lootScroll.buttons) do
  item:HookScript("OnEnter", AddPlayersGameTooltip)
end

hooksecurefunc(EncounterJournal.encounter.info.lootScroll, "update", function()
	local items = EncounterJournal.encounter.info.lootScroll.buttons;
  for _, item in pairs(items) do
    if item.showingTooltip then
      AddPlayersGameTooltip(item)
    end
  end
end)